#!/usr/bin/env python2
# -*- coding: utf-8 -*-

r""" Billard simulation tools """


from __future__ import (unicode_literals, absolute_import, division,
                        print_function)

# To ignore numpy errors:
#pylint: disable=no-name-in-module
from numpy import (array, Inf, linspace, pi, cos, sin)
#pylint: enable=no-name-in-module
from numpy.linalg import norm
from shapely.geometry import Polygon, Point, LineString
from itertools import izip


### Here we monkey-patch the Point class to add the __add__ and __mul__
### operators
def _add_points(self, other):
    r""" Add two Points """
    result = array(self) + array(other)
    return Point(result)

def _mul_points(self, coef):
    r""" Multiply a point by coef """
    result = array(self) * coef
    return Point(result)

Point.__add__ = _add_points
Point.__mul__ = _mul_points
Point.__rmul__ = _mul_points

class BilliardBadDirection(ValueError):
    pass

class Billard(object):
    r""" Construct a billard from a polygon """

    def __init__(self, polygon):
        r""" Construct a billiard from a polygon """
        self.polygon = polygon
        self.distances_to_vertex = {}
        for (xpos, ypos) in izip(*self.boundary.xy):
            point = Point(xpos, ypos)
            dist = self.boundary.project(point)
            self.distances_to_vertex[dist] = point

        #make it periodic
        self.distances_to_vertex[self.length] = self.distances_to_vertex[0]

        self.vertex_distances = array(self.distances_to_vertex.keys())

    @property
    def boundary(self):
        r""" return the billard boundary """
        return self.polygon.boundary

    @property
    def length(self):
        r""" return the billard boundary length (perimeter) """
        return self.polygon.length

    def contains(self, other):
        r""" return True if `other` is contained in the billard geometry """
        return self.polygon.contains(other)

    def touches(self, other):
        r""" return True if `other` is touching in the billard geometry """
        return self.polygon.touches(other)

    def collision(self, point, direction):
        r""" Return the point and the new direction, where the collision between
        the given point, having the given direction, takes place """

        #This is the algorithm :
        #  From the original point p, we construct a line, having the direction
        #  `direction`. Then, we find the intersection points between this line
        #  and the billard and we choose the closest one (different of p).

        line = LineString([array(point), array(point) + self.length*direction])

        candidates = self.boundary.intersection(line)
        dist_min = Inf
        collision_point = None
        try:
            for candidate in candidates:
                distance = candidate.distance(point)
                if distance < dist_min and not candidate.almost_equals(point):
                    dist_min = distance
                    collision_point = candidate
        except TypeError: #not iterable, i.e. there is only one intersection
            collision_point = candidates

        if collision_point is None:
            raise BilliardBadDirection()

        #now we have the colliding point, we have to compute the new direction.
        #to do that, we have to find the edge extrimities, called A and B

        #each point of the boundary is known by its distance from the initial
        #point. We first compute the distance from 0 of the colliding point

        collision_dist = self.boundary.project(collision_point)

        #just assert the point touches the billard
        if not self.touches(collision_point):
            collision_point = self.boundary.interpolate(collision_dist)
            collision_dist = self.boundary.project(collision_point)

        #here we compute the difference between the collision_dist and the
        #vertex distances. The points before will be negative while the other
        #positive. The challenge is to find the edges before and after the
        #collision.

# Here an example with a simple billard. The `v` means the collision point
#
#      C x-------v---x D
#        |           |
#      B x-----------x A
#
#       vertex_distances = [0, 1, 2, 3]
#       collision_dist = 2.7
#
#       v_min = [-2.7, -1.7, -0.7, -Inf] # where v_min > 0, we set -Inf
#       v_max = [Inf, Inf, Inf, 0.3] # where v_max < 0, we set +Inf
#
#       the argmax of v_min gives the point C
#       the argmin of v_max gives the point D.


        v_min = array(self.vertex_distances - collision_dist)
        v_max = array(self.vertex_distances - collision_dist)

        v_min[v_min >= 0] = -Inf
        v_max[v_max <= 0] = Inf

        a_index = v_min.argmax()
        b_index = v_max.argmin()

        point_a = self.distances_to_vertex[self.vertex_distances[a_index]]
        point_b = self.distances_to_vertex[self.vertex_distances[b_index]]

        #knowning the two extrimities we compute the normal vector
        normal_vector = array([point_b.y - point_a.y, point_a.x - point_b.x])
        normal_vector /= norm(normal_vector)

        #assert normal_vector is outward
        diff_x = 1e-4
        if self.polygon.contains(collision_point + diff_x*normal_vector):
            normal_vector *= -1

        #finally we compute the new direction.
        projection = (direction*normal_vector).sum()
        new_direction = direction - 2*projection*normal_vector
        new_direction /= norm(new_direction)


        #for the Poincarré map, one needs the normed collision dist and
        #the value of sin(\phi), where \phi is the angle between the direction
        #and the normal vector
        distance = collision_dist / self.length
        sinphi = direction[1]*normal_vector[0] - direction[0]*normal_vector[1]
        sinphi /= norm(direction)

        return collision_point, new_direction, sinphi, distance

#A list of billard builders

#pylint: disable=invalid-name
#pylint: disable=missing-docstring
def _build_limacon(param):
    a = param.get('a', 1)

    theta = linspace(0, 2*pi, N_POINTS, endpoint=False)
    rho = a + cos(theta)

    x = rho*cos(theta)
    y = rho*sin(theta)
    limacon = Polygon(zip(x, y))
    return Billard(limacon)

def _build_sinai(param):
    a = param.get('a', 2)
    b = param.get('b', 1)
    theta = linspace(0, 2*pi, N_POINTS, endpoint=False)

    rectangle = Polygon([[-a, -b], [-a, b], [a, b], [a, -b]])
    x = a + b*cos(theta)
    y = b + b*sin(theta)
    circle = Polygon(zip(x, y))

    sinai = rectangle.difference(circle)
    return Billard(sinai)


def _build_circle(param):
    r = param.get('radius', 1)
    return _build_elliptic({'a':r, 'b':r})

def _build_square(param):
    l = param.get('length', 1)
    square = Polygon([[-l, -l], [-l, l], [l, l], [l, -l]])
    return Billard(square)


def _build_stadium(param):
    a = param.get('a', 2)
    b = param.get('b', 1)
    theta = linspace(pi/2, -pi/2, N_POINTS/2, endpoint=False)

    edges = [[-a, b], [a, b]]
    x = a + b*cos(theta)
    y = 0 + b*sin(theta)
    edges.extend(zip(x, y))
    edges.extend([[a, -b], [-a, -b]])
    edges.extend(zip(-x, -y))

    stadium = Polygon(edges)
    return Billard(stadium)


def _build_elliptic(param):
    a = param.get('a', 2)
    b = param.get('b', 1)

    theta = linspace(0, 2*pi, N_POINTS, endpoint=False)
    x = a*cos(theta)
    y = b*sin(theta)

    ellipse = Polygon(zip(x, y))
    return Billard(ellipse)

#pylint: enable=invalid-name
#pylint: enable=missing-docstring

BILLARDS_BUILDERS = {'limacon': _build_limacon,
                     'sinai': _build_sinai,
                     'circle': _build_circle,
                     'square': _build_square,
                     'stadium': _build_stadium,
                     'elliptic': _build_elliptic
                    }
SUPPORTED_BILLARDS = BILLARDS_BUILDERS.keys()
N_POINTS = 150

def buildbillard(name, ownbuilder=None, **param):
    try:
        return ownbuilder(param)
    except TypeError:
        try:
            return BILLARDS_BUILDERS[name.lower()](param)
        except KeyError:
            raise ValueError('builder not found')

def main():
    from shapely_plot import plot_shape
#pylint: disable=no-name-in-module
    from pylab import (rand, random, pause, xlim, ylim, gca, subplot, clf)
#pylint: enable=no-name-in-module

    while True:
        for num, billard_name in enumerate(SUPPORTED_BILLARDS):
            print('{}) {}'.format(num, billard_name))
        wanted = raw_input('Which billard do you want ? (negative number to quit) : ')
        wanted = int(wanted)
        if wanted < 0:
            break
        try:
            billard = buildbillard(SUPPORTED_BILLARDS[wanted])
        except IndexError:
            continue

        point = billard.boundary.interpolate(random(), normalized=True)
        direction = rand(1, 2)[0]
        direction /= norm(direction)

        clf()
        subplot(211)
        ax = gca()
        plot_shape(billard.boundary, ax, 'b').plot
        plot_shape(point, ax, 'r').plot

        subplot(212)
        bx = gca()
        xlim(-0.1, 1.1)
        ylim(-1.1, 1.1)

        for i in xrange(100):
            np, nd, phi, l = billard.collision(point, direction)
            direction_line = LineString([array(point).tolist(), array(np).tolist()])

            bx.plot(l, phi, 'b.')

            plot_shape(np, ax, 'r').plot
            plot_shape(direction_line, ax, 'g').plot
            pause(0.0001)

            point = np
            direction = nd

if __name__ == '__main__':
    main()
