#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from __future__ import (unicode_literals, absolute_import, division,
                        print_function)

from PySide import QtGui
from PySide.QtCore import Slot, Qt

from threading import Thread

import os
os.environ["QT_API"] = "pyside"

import matplotlib
matplotlib.use('Qt4Agg')
from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
matplotlib.rcParams['backend.qt4']='PySide'
import matplotlib.pyplot as plt

from numpy import array, pi, cos, sin, Inf
from numpy.linalg import norm
from pylab import rand, random
from shapely.geometry import LineString
from random import choice

from mainwindow_ui import Ui_MainWindow

from shapely.geometry import Polygon
from billard import (SUPPORTED_BILLARDS, buildbillard, Billard,
                     BilliardBadDirection)
from shapely_plot import plot_shape

COLORS = set('gcmyk')

class MainWindow(QtGui.QMainWindow):
    def __init__(self):
        super(MainWindow, self).__init__()
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        self._init_plot()
        self._init_list_of_billards()

        self.running = False
        self.ui.run_button.clicked.connect(self.toggle)
        self.ui.position_slider.valueChanged.connect(self.change_initial_position)
        self.ui.direction_slider.valueChanged.connect(self.change_initial_position)
        self.show()

    def _init_plot(self):
        self.figure = plt.figure()
        self.canvas = FigureCanvas(self.figure)
        self.ui.main_layout.addWidget(self.canvas)

    def _init_list_of_billards(self):
        self.billard_list = ['NONE', 'CUSTOM'] + SUPPORTED_BILLARDS
        self.ui.billard_selection.addItems(self.billard_list)
        self.ui.billard_selection.currentIndexChanged.connect(self.change_billard)

    def closeEvent(self, event):
        self.running = False #to kill the thread
        QtGui.QMainWindow.closeEvent(self, event)

    def new_color(self):
        try:
            self.color = choice(tuple(COLORS - set(self.color)))
        except AttributeError:
            self.color = choice(tuple(COLORS))

    def set_counter(self, value):
        self.ui.iteration_counter.setValue(value)

    def get_counter(self):
        return self.ui.iteration_counter.value()

    counter = property(get_counter, set_counter)

    @Slot()
    def start(self):
        self.running = True
        self.billard_thread = Thread(target=self._nextstep)
        self.billard_thread.start()
        self.ui.run_button.setText('stop')
        self.new_color()

    @Slot()
    def stop(self):
        self.running = False
        self.ui.run_button.setText('start')

    @Slot()
    def toggle(self):
        if self.running:
            self.stop()
        else:
            self.start()

    def set_buttons_enabled(self, enabled=True):
        self.ui.run_button.setEnabled(enabled)
        self.ui.position_slider.setEnabled(enabled)
        self.ui.direction_slider.setEnabled(enabled)

    def _nextstep(self):
        mod = 10
        counter = 0
        while self.running:
            p = self.point
            direction = self.direction
            try:
                np, nd, phi, l = self.billard.collision(p, direction)
            except BilliardBadDirection:
                angle = (self.ui.direction_slider.value() + 180) % 360
                self.ui.direction_slider.setValue(angle)
            direction = LineString([array(p).tolist(), array(np).tolist()])

            self.poincarre_plot.plot(l, phi, '{}.'.format(self.color), markersize=4)

            if self.counter < 100:
                plot_shape(direction, self.billard_plot, self.color).plot

            self.point = np
            self.direction = nd
            self.canvas.draw()
            counter += 1
            if counter % mod == 0:
                self.counter += mod

    @Slot(int)
    def change_billard(self, index, ownbuilder=None):
        billard_name = self.billard_list[index]

        if billard_name == 'NONE':
            self.set_buttons_enabled(False)
            self.stop()
            return

        if billard_name == 'CUSTOM' and not ownbuilder:
            return self.make_custom_billard()

        self.set_buttons_enabled(True)
        self.billard = buildbillard(billard_name, ownbuilder)
        self.reset()
        self.change_initial_position() #called it once, at least

    def make_custom_billard(self):
        self.figure.clf()

        def clear_plot():
            plot.cla()
            plot.set_xlim(-1, 1)
            plot.set_ylim(-1, 1)

        plot = self.figure.add_subplot(111)
        clear_plot()

        positions = []
        dist = Inf
        min_dist = 1e-1

        def onclick(event):
            if event.button != 1 or event.inaxes is None:
                return
            positions.append(array([event.xdata, event.ydata]))
            clear_plot()
            a = array(positions)
            plot.plot(a[:,0], a[:,1], 'b--')
            plot.plot(a[:,0], a[:,1], 'r+', markersize=20)
            self.canvas.draw()

            dist = norm(positions[-1] - positions[0])
            if len(positions) > 2 and dist < min_dist:
                def builder(param):
                    return Billard(Polygon(positions))
                #the last position can be remove, as the Polygon function will
                #automatically find the good point to close it.
                positions.pop()
                return self.change_billard(1, builder) #fix index 1

        self.figure.canvas.mpl_connect('button_press_event', onclick)
        self.figure.tight_layout()
        self.canvas.draw()

    def reset(self):
        self.counter = 0
        self.reset_plot()

    def reset_plot(self):
        self.figure.clf()
        self.billard_plot = self.figure.add_subplot(211)
        plot_shape(self.billard.boundary, self.billard_plot, 'b').plot

        self.poincarre_plot = self.figure.add_subplot(212)
        self.poincarre_plot.set_xlim(-0.1, 1.1)
        self.poincarre_plot.set_ylim(-1.1, 1.1)

        self.point_plot = None

        self.figure.tight_layout()
        self.canvas.draw()

    @Slot()
    def change_initial_position(self):
        self.stop()

        position = self.ui.position_slider.value()/100
        angle = (270 - self.ui.direction_slider.value() % 360)*pi/180

        self.point = self.billard.boundary.interpolate(position, normalized=True)
        self.direction = array([cos(angle), sin(angle)])

        x, y = self.point.x, self.point.y
        try:
            self.point_plot.set_data(x, y)
        except AttributeError:
            self.reset_plot()
            self.point_plot = self.billard_plot.plot(x, y, 'r.', markersize=20)[0]
        self.canvas.draw()
